
//		CEsc.cpp - expand C-style escape sequences, or escapify text


#define UNICODE					1
#define _UNICODE				1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>

#include "PlugIn.h"
#include "TakeCmd.h"


#define VER_MAJOR				0
#define VER_MINOR				60
#define VER_BUILD				0
#define VER_PATCH				5

#define PLUGIN_NAME				L"CEsc"

#ifdef _WIN64
#define PLATFORM				L"x64"
#else
#define PLATFORM				L"x86"
#endif

#define MAX_ARG_LEN				8192
#define SHORT_BUF_LEN			1024
#define MAX_FILENAME_LEN		32768

#define ERRHANDLE GetStdHandle( STD_ERROR_HANDLE )


#define MACRO_EXP( arg )		#arg
#define MACRO_STR( arg )		MACRO_EXP( arg )


HINSTANCE MyDLLHandle			= NULL;


const wchar_t PluginFeatures[] = L"@escapify,@escenv,@escline,@escread,@exp,@expcmp,@expfind,@expreplace,@expwrite,_cescver,expecho";



const wchar_t ExpEchoHelpText[]		= L"Expand escapes and display text.\n\
\n\
EXPECHO text\r\n\
\n\
Escape sequences will be expanded, and the resulting text \n\
sent to stdout.  No end-of-line will be appended. \n";


#ifdef CESC_EXPTEST_CMD

const wchar_t ExpTestHelpText[]		= L"Test expansion.\n\
\n\
EXPTEST text\r\n\
\n\
Escape sequences will be expanded, and the resulting text \n\
sent to stdout. \n";

#endif	//	CESC_EXPTEST_CMD


const wchar_t EscCharsVarName[]		= L"CEsc_Chars";

const wchar_t EscFlagsVarName[]		= L"CEsc_Flags";

const wchar_t UnsafeChars[]			= L"\"%&<>[]^`|";

const wchar_t MinusOneStr[]			= L"-1";


wchar_t svSetDos[32] = L"";



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



int GetErrorText( int err, LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, err, 0, outString, SHORT_BUF_LEN, NULL ) == 0 )
		wsprintf( outString, L"Error %d", err );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



int GetLastErrorText( LPTSTR outString )
{
	return GetErrorText( GetLastError(), outString );
}



//	Calls an internal TCC command.

int CallInternal( LPTSTR Cmd, LPTSTR Args )
{
	if ( !Cmd )
		return -1;
	else if ( Cmd[0] == '\0' )
		return -1;


	wchar_t CmdBuf[MAX_ARG_LEN]		= L"";
	wchar_t ExpVar[SHORT_BUF_LEN]	= L"%_expansion";
	BOOL AddStar					= TRUE;


	ExpandVariables( ExpVar, 1 );
	if ( wcschr( ExpVar, '1' ) )
		AddStar = FALSE;


	if ( ( Cmd[0] != '*' ) && AddStar )
		wcscpy_s( CmdBuf, MAX_ARG_LEN, L"*" );

	wcscat_s( CmdBuf, MAX_ARG_LEN, Cmd );

	if ( Args ) {
		if ( Args[0] ) {
			if ( !iswspace( Args[0] ) )
				wcscat_s( CmdBuf, MAX_ARG_LEN, L" " );

			wcscat_s( CmdBuf, MAX_ARG_LEN, Args );
		}
	}

	return Command( CmdBuf, 0 );
}



#include "ParseArgs.cpp"
#define EXPAND_ESCAPES_REAL_BACKSPACE
#include "ExpandEscapes.cpp"
#include "NewHelp.cpp"



BOOL IsColonOrEquals( wchar_t ch )
{
	if ( ( ch == ':' ) || ( ch == '=' ) )
		return TRUE;
	else
		return FALSE;
}



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned int value	= 0;
	wchar_t *sep1		= NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return errno;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return errno;
		else if ( !EndChar && *sep1 )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



int ParseSignedInt( LPTSTR inString, int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	int value		= 0;
	wchar_t *sep1	= NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return errno;
		else if ( *sep1 && !EndChar )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) || ( ( inString[0] == '-' ) && iswdigit( inString[1] ) ) ) {
		//  parse signed integer

		value = wcstol( inString, &sep1, 10 );

		if ( errno )
			return errno;
		else if ( *sep1 && !EndChar )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



int EvaluateQuietly( LPTSTR String )
{
	HANDLE OldStderrHandle	= GetStdHandle( STD_ERROR_HANDLE );
	HANDLE OldStdoutHandle	= GetStdHandle( STD_OUTPUT_HANDLE );
	HANDLE NewStderrHandle	= CreateFile( L"nul", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_ALWAYS, 0, NULL );
	int rv		= 0;


	if ( NewStderrHandle != INVALID_HANDLE_VALUE ) {
		SetStdHandle( STD_ERROR_HANDLE, NewStderrHandle );		//  redirect stderr to nul:
		SetStdHandle( STD_OUTPUT_HANDLE, NewStderrHandle );		//  redirect stdout to nul:
	}
	
	rv = Evaluate( String );									//  and evaluate the expression

	if ( NewStderrHandle != INVALID_HANDLE_VALUE ) {
		SetStdHandle( STD_ERROR_HANDLE, OldStderrHandle );		//  put stderr back where we found it
		SetStdHandle( STD_OUTPUT_HANDLE , OldStdoutHandle );	//  put stdout back where we found it
		CloseHandle( NewStderrHandle );
	}

	return rv;
}



int AppendFilename( LPTSTR Path, LPTSTR Filename, size_t BufferSize )
{
	if ( !Path || !Filename )
		return -1;

	
	BOOL NeedSlash		= FALSE;

	for ( size_t i = 0; Path[i]; i++ ) {
		if ( ( Path[i] == '/' ) || ( Path[i] == '\\' ) || ( Path[i] == ':' ) )
			NeedSlash = FALSE;
		else
			NeedSlash = TRUE;
	}

	if ( ( Filename[0] == '/' ) || ( Filename[0] == '\\' ) )
		NeedSlash = FALSE;
	
	if ( NeedSlash )
		wcscat_s( Path, BufferSize, L"\\" );

	wcscat_s( Path, BufferSize, Filename );
	return 0;
}



//	Strips off any leading or trailing spaces from a string.

int TrimLine( LPTSTR String )
{
	if ( !String )
		return -1;
	else if ( String[0] == '\0' )
		return 0;

	int FirstNonSpace = -1, LastNonSpace = -1;
	wchar_t ch = '\0';

	for ( int i = 0; String[i]; i++ ) {
		ch = String[i];

		if ( !iswspace( ch ) ) {
			LastNonSpace = i;

			if ( FirstNonSpace == -1 )
				FirstNonSpace = i;
		}
	}

	if ( FirstNonSpace == -1 ) {
		String[0] = '\0';
		return 0;
	}

	String[LastNonSpace + 1] = '\0';

	if ( FirstNonSpace ) {
		int i = 0;

		do {
			ch = String[FirstNonSpace + i];
			String[i++] = ch;

		} while ( ch );
	}

	return 0;
}



wchar_t *FirstFoundChar( LPTSTR String, LPCTSTR SearchChars )
{
	if ( !String || !SearchChars )
		return NULL;
	else if ( ( String[0] == '\0' ) || ( SearchChars[0] == '\0' ) )
		return NULL;


	wchar_t *ch			= String;

	while ( *ch ) {
		if ( wcschr( SearchChars, *ch ) )
			return ch;

		ch++;
	}

	return NULL;
}



wchar_t GetCommandSep()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"%@option[commandsep]";


	ExpandVariables( temp, 1 );

	if ( temp[0] && !temp[1] )
		return temp[0];

	return '&';
}



wchar_t GetSynEscChar()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"%@option[escapechar]";


	ExpandVariables( temp, 1 );

	if ( temp[0] && !temp[1] )
		return temp[0];

	return '^';
}



int GetEscChars( LPTSTR Buffer, size_t buflen, unsigned int *outFlags )
{
	if ( !Buffer || !buflen )
		return -1;


	wchar_t temp[SHORT_BUF_LEN]		= L"";

	wchar_t CommandSep	= GetCommandSep();
	wchar_t SynEscChar	= GetSynEscChar();
	unsigned int t		= 0;
	size_t len			= 0;
	wchar_t *sep1		= NULL;


	wcscpy_s( Buffer, buflen, UnsafeChars );
	len = wcslen( Buffer );

	if ( CommandSep ) {
		if ( !wcschr( Buffer, CommandSep ) ) {
			if ( len < ( buflen - 1 ) ) {
				Buffer[len++]	= CommandSep;
				Buffer[len]		= '\0';
			}
		}
	}

	if ( SynEscChar ) {
		if ( !wcschr( Buffer, SynEscChar ) ) {
			if ( len < ( buflen - 1 ) ) {
				Buffer[len++]	= SynEscChar;
				Buffer[len]		= '\0';
			}
		}
	}

	GetEnvironmentVariable( EscCharsVarName, Buffer + len, (DWORD) ( buflen - len ) );

	if ( outFlags ) {
		temp[0] = '\0';
		GetEnvironmentVariable( EscFlagsVarName, temp, SHORT_BUF_LEN );

		if ( iswdigit( temp[0] ) ) {
			if ( ParseInt( temp, &t, &sep1 ) )
				t = 0;
			else if ( *sep1 )
				t = 0;
		}

		if ( temp[0] )
			*outFlags = t;
	}

	return 0;
}



int PrepSetDos()
{
	wcscpy_s( svSetDos, 32, L"%_expansion" );

	ExpandVariables( svSetDos, 1 );
	if ( svSetDos[0] == '\0' )
		wcscpy_s( svSetDos, 32, L"0" );

	CallInternal( L"setdos", L" /x-1478+3" );
	return 0;
}



int ResetSetDos()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"/X";


	wcscat_s( temp, SHORT_BUF_LEN, svSetDos );

	CallInternal( L"setdos", temp );
	return 0;
}



DLLExports int WINAPI f_escapify( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t EscChars[SHORT_BUF_LEN]		= L"";

	unsigned int EscFlags				= 0;


	GetEscChars( EscChars, SHORT_BUF_LEN, &EscFlags );

	if ( Escapify( String, EscFlags, EscChars, MAX_ARG_LEN ) == ESC_BUFFER_OVERFLOW ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Buffer overflow\r\n" );
		return -1;
	}

	return 0;
}



DLLExports int WINAPI f_exp( LPTSTR String )
{
	if ( !String )
		return 1;


	if ( ExpandEscapes( String, MAX_ARG_LEN ) == ESC_BUFFER_OVERFLOW ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Buffer overflow\r\n" );
		return -1;
	}

	return 0;
}



DLLExports int WINAPI f_escline( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t temp[MAX_ARG_LEN]			= L"%@line[";
	wchar_t EscChars[SHORT_BUF_LEN]		= L"";

	unsigned int EscFlags			= 0;
	int rv							= 0;


	wcscat_s( temp, MAX_ARG_LEN, String );
	wcscat_s( temp, MAX_ARG_LEN, L"]" );

	PrepSetDos();
	rv = ExpandVariables( temp, 1 );
	ResetSetDos();

	if ( rv )
		return rv;


	GetEscChars( EscChars, SHORT_BUF_LEN, &EscFlags );

	wcscpy_s( String, MAX_ARG_LEN, temp );

	if ( Escapify( String, EscFlags, EscChars, MAX_ARG_LEN ) == ESC_BUFFER_OVERFLOW ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Buffer overflow\r\n" );
		return -1;
	}

	return 0;
}



DLLExports int WINAPI f_escread( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t temp[MAX_ARG_LEN]			= L"%@fileread[";
	wchar_t EscChars[SHORT_BUF_LEN]		= L"";

	unsigned int EscFlags		= 0;
	int rv						= 0;


	wcscat_s( temp, MAX_ARG_LEN, String );
	wcscat_s( temp, MAX_ARG_LEN, L"]" );

	PrepSetDos();
	rv = ExpandVariables( temp, 1 );
	ResetSetDos();

	if ( rv )
		return rv;


	GetEscChars( EscChars, SHORT_BUF_LEN, &EscFlags );

	wcscpy_s( String, MAX_ARG_LEN, temp );

	if ( Escapify( String, EscFlags, EscChars, MAX_ARG_LEN ) == ESC_BUFFER_OVERFLOW ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Buffer overflow\r\n" );
		return -1;
	}
	
	return 0;
}



DLLExports int WINAPI f_escenv( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t temp[MAX_ARG_LEN]			= L"";
	wchar_t EscChars[SHORT_BUF_LEN]		= L"";

	LPTSTR VarName					= String;
	unsigned int EscFlags			= 0;


	TrimLine( String );

	while ( VarName[0] == '%' )
		VarName++;

	if ( VarName[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No variable name\r\n" );
		return -1;
	}
	else if ( !iswalpha( VarName[0] ) && ( VarName[0] != '$' ) && ( VarName[0] != '_' ) ) {
		DisplayErrorHeader();
		_wcsupr_s( String, MAX_ARG_LEN );
		Qprintf( ERRHANDLE, L"Bad variable name: \"%s\"\r\n", VarName );
		return -1;
	}


	GetEscChars( EscChars, SHORT_BUF_LEN, &EscFlags );

	GetEnvironmentVariable( VarName, temp, MAX_ARG_LEN );

	if ( Escapify( temp, EscFlags, EscChars, MAX_ARG_LEN ) == ESC_BUFFER_OVERFLOW ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Buffer overflow\r\n" );
		return -1;
	}

	wcscpy_s( String, MAX_ARG_LEN, temp );
	return 0;
}



unsigned int GetOutputEncoding()
{
	wchar_t temp[SHORT_BUF_LEN]			= L"%@option[utf8output]";

	unsigned int CodePage				= 0;


	if ( QueryUnicodeOutput() )
		return 0;

	ExpandVariables( temp, 1 );
	if ( towupper( temp[0] ) == 'Y' )
		return CP_UTF8;

	wcscpy_s( temp, SHORT_BUF_LEN, L"%_codepage" );
	ExpandVariables( temp, 1 );

	if ( iswdigit( temp[0] ) )
		CodePage = wcstoul( temp, NULL, 10 );

	if ( !CodePage )
		CodePage = GetOEMCP();

	return CodePage;
}



int WriteBomIfNeeded( HANDLE Handle, unsigned int Encoding )
{
	if ( Encoding && ( Encoding != CP_UTF8 ) )
		return TRUE;


	BYTE Bom[2][3]			= { { 0xff, 0xfe, 0x00 }, { 0xef, 0xbb, 0xbf } };

	BY_HANDLE_FILE_INFORMATION fi;
	DWORD BytesWritten		= 0;
	int rv					= TRUE;


	if ( GetFileInformationByHandle( Handle, &fi ) ) {
		if ( ( fi.nFileSizeHigh == 0 ) && ( fi.nFileSizeLow == 0 ) ) {

			switch ( Encoding ) {

			case CP_UTF8 :

				rv = WriteFile( Handle, Bom[1], 3, &BytesWritten, NULL );
				break;

			default :

				rv = WriteFile( Handle, Bom[0], 2, &BytesWritten, NULL );
			}
		}
	}

	return rv;
}



DLLExports int WINAPI f_expwrite( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t temp[MAX_ARG_LEN]			= L"";

	HANDLE Handle					= NULL;
	wchar_t *ch						= String;
	wchar_t *sep1					= NULL;
	unsigned int OutputEncoding		= 0;
	size_t NumChars					= 0;
	DWORD BytesWritten				= 0;
	int rv							= 0;


	while ( iswspace( *ch ) )
		ch++;

	if ( *ch == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing handle\r\n" );
		return -1;
	}
	else if ( ( *ch == '-' ) && ( ( ch[1] == ',' ) || ( ch[1] == ';' ) || iswspace( ch[1] ) ) ) {
		Handle = GetStdHandle( STD_OUTPUT_HANDLE );
		sep1 = ch + 1;
	}
	else if ( ParseInt( ch, (unsigned int*) &Handle, &sep1 ) ) {
		sep1 = FirstFoundChar( ch, L",;" );
		if ( sep1 )
			*sep1 = '\0';

		_wcsupr_s( ch, MAX_ARG_LEN );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad handle: \"%s\"\r\n", ch );
		return -1;
	}
	else if ( !Handle ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Bad handle: \"%u\"\r\n", Handle );
		return -1;
	}

	while ( iswspace( *sep1 ) )
		sep1++;

	if ( ( *sep1 != ',' ) && ( *sep1 != ';' ) ) {
		_wcsupr_s( ch, MAX_ARG_LEN );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Expected a comma: \"%s\"\r\n", ch );
		return -1;
	}


	sep1++;
	wcscpy_s( temp, MAX_ARG_LEN, sep1 );



	if ( ExpandEscapes( temp, MAX_ARG_LEN ) == ESC_BUFFER_OVERFLOW ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Buffer overflow\r\n" );
		return -1;
	}
	else if ( wcslen( temp ) >= ( MAX_ARG_LEN - 2 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Buffer overflow\r\n" );
		return -1;
	}

	wcscat_s( temp, MAX_ARG_LEN, L"\r\n" );
	NumChars = wcslen( temp );


	OutputEncoding = GetOutputEncoding();

	if ( !WriteBomIfNeeded( Handle, OutputEncoding ) ) {
		wcscpy_s( String, MAX_ARG_LEN, MinusOneStr );
		return 0;
	}


	if ( OutputEncoding == 0 ) {		//	UTF-16
		rv = WriteFile( Handle, temp, (DWORD) ( NumChars * sizeof( wchar_t ) ), &BytesWritten, NULL );
	}
	else {								//	UTF-8 or OEM
		char OutBuffer[4 * MAX_ARG_LEN]		= "";

		WideCharToMultiByte( OutputEncoding, 0, temp, -1, OutBuffer, 4 * MAX_ARG_LEN, NULL, NULL );
		NumChars = strlen( OutBuffer );
		rv = WriteFile( Handle, OutBuffer, (DWORD) NumChars, &BytesWritten, NULL );
	}


	if ( rv )
		wsprintf( String, L"%u", NumChars );
	else
		wcscpy_s( String, MAX_ARG_LEN, MinusOneStr );

	return 0;
}



DLLExports int WINAPI f_expcmp( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t String1[MAX_ARG_LEN]			= L"";
	wchar_t String2[MAX_ARG_LEN]			= L"";

	unsigned int Flags		= 0;
	int rv					= 0;


	ParseArgs( String, PARSE_BREAK_COMMAS );


	if ( arg[0] ) {
		wcscpy_s( String1, MAX_ARG_LEN, arg[0] );

		if ( ExpandEscapes( String1, MAX_ARG_LEN ) == ESC_BUFFER_OVERFLOW ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Buffer overflow expanding string1\r\n" );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing string1\r\n" );
		return -1;
	}

	if ( arg[1] ) {
		wcscpy_s( String2, MAX_ARG_LEN, arg[1] );

		if ( ExpandEscapes( String2, MAX_ARG_LEN ) == ESC_BUFFER_OVERFLOW ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Buffer overflow expanding string2\r\n" );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing string2\r\n" );
		return -1;
	}

	if ( arg[2] ) {
		if ( ParseInt( arg[2], &Flags, NULL ) ) {
			_wcsupr_s( arg[2], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad flags: \"%s\"\r\n", arg[2] );
			return -1;
		}
	}


	if ( Flags & 1 ) {
		CharLowerW( String1 );
		CharLowerW( String2 );
	}

	rv = wcscmp( String1, String2 );

	wsprintf( String, L"%d", rv );
	return 0;
}



unsigned int Exp_GetExpandedOffset( unsigned int OriginalOffset, unsigned int OriginalLen, unsigned int ExpandedLen, DWORD *CharFrom )
{
	if ( !OriginalOffset || !OriginalLen || !ExpandedLen || !CharFrom )
		return 0;


	if ( OriginalOffset >= OriginalLen )
		return ExpandedLen;

	for ( unsigned int i = 1; i < ExpandedLen; i++ ) {
		if ( CharFrom[i] >= OriginalOffset )
			return i;
	}

	return ExpandedLen;
}



DLLExports int WINAPI f_expfind( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t String1[MAX_ARG_LEN]			= L"";
	wchar_t Substring[MAX_ARG_LEN]			= L"";

	DWORD CharFrom[MAX_ARG_LEN];
	unsigned int Flags			= 0;
	int n						= 1;
	unsigned int offset			= 0;

	wchar_t *Found				= NULL;
	unsigned int Matches		= 0;
	wchar_t **MatchN			= NULL;
	unsigned int offset1		= 0;

	size_t l1					= 0;
	size_t l2					= 0;
	size_t i					= 0;
	size_t j					= 0;


	ParseArgs( String, PARSE_BREAK_COMMAS );


	if ( arg[0] ) {
		wcscpy_s( String1, MAX_ARG_LEN, arg[0] );

		if ( ExpandEscapes( String1, MAX_ARG_LEN, CharFrom ) == ESC_BUFFER_OVERFLOW ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Buffer overflow expanding string\r\n" );
			return -1;
		}
	}

	if ( arg[1] ) {
		wcscpy_s( Substring, MAX_ARG_LEN, arg[1] );

		if ( ExpandEscapes( Substring, MAX_ARG_LEN ) == ESC_BUFFER_OVERFLOW ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Buffer overflow expanding substring\r\n" );
			return -1;
		}
	}

	if ( arg[2] ) {
		if ( ParseSignedInt( arg[2], &n, NULL ) ) {
			_wcsupr_s( arg[2], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad n: \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( arg[3] ) {
		if ( ParseInt( arg[3], &Flags, NULL ) ) {
			_wcsupr_s( arg[3], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad flags: \"%s\"\r\n", arg[3] );
			return -1;
		}
	}

	if ( arg[4] ) {
		wchar_t temp[MAX_ARG_LEN]		= L"";


		wcscpy_s( temp, MAX_ARG_LEN, arg[4] );
		EvaluateQuietly( temp );

		if ( ParseInt( temp, &offset, NULL ) ) {
			_wcsupr_s( arg[4], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad offset: \"%s\"\r\n", arg[4] );
			return -1;
		}
	}


	if ( ( Flags & 1 ) == 0 ) {
		CharLowerW( String1 );
		CharLowerW( Substring );
	}

	l1 = wcslen( String1 );
	l2 = wcslen( Substring );

	offset1 = Exp_GetExpandedOffset( offset, arglen[0], (unsigned int) l1, CharFrom );
	//	wprintf( L"   *  offset = %u    offset1 = %u\n", offset, offset1 );


	//		count matches :

	Found = String1 + offset1;

	if ( String1[offset1] && Substring[0] ) {
		do {
			Found = wcsstr( Found, Substring );

			if ( Found ) {
				Matches++;
				Found++;
			}
		}
		while ( Found );
	}


	if ( !n ) {
		wsprintf( String, L"%u", Matches );
		return 0;
	}


	//		store list of all matches in array :

	if ( Matches ) {

		MatchN = (wchar_t**) calloc( Matches, sizeof( wchar_t* ) );
		if ( !MatchN ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error allocating matches buffer\r\n" );
			return -1;
		}

		Found = String1 + offset1;

		for ( unsigned int k = 0; k < Matches; k++ ) {
			MatchN[k] = wcsstr( Found, Substring );

			if ( MatchN[k] == NULL ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Internal error\r\n" );
				free( MatchN );
				return -1;
			}

			Found = MatchN[k] + 1;
		}
	}


	Found = NULL;

	if ( n > 0 ) {
		if ( (unsigned int) n <= Matches )
			Found = MatchN[n - 1];
	}
	else {
		if ( (unsigned int) -n <= Matches )
			Found = MatchN[Matches + n];
	}

	if ( Found ) {
		i = Found - String1;

		if ( ( i + l2 ) >= l1 )
			j = (unsigned int) arglen[0] - CharFrom[i];
		else
			j = CharFrom[i + l2] - CharFrom[i];

		if ( Flags & 8 )
			wsprintf( String, L"%u+%u", CharFrom[i], j );
		else
			wsprintf( String, L"%u", CharFrom[i] );
	}
	else
		wcscpy_s( String, MAX_ARG_LEN, MinusOneStr );

	free( MatchN );
	return 0;
}



wchar_t *WcsIStr( const wchar_t *String, const wchar_t *SubString )
{
	if ( !String || !SubString )
		return NULL;

	if ( ( String[0] == '\0' ) || ( SubString[0] == '\0' ) )
		return NULL;


	wchar_t *StringLower		= NULL;
	wchar_t *SubStringLower		= NULL;
	wchar_t *Found				= NULL;
	size_t len1					= wcslen( String ) + 1;
	size_t len2					= wcslen( SubString ) + 1;


	if ( len2 > len1 )
		return NULL;

	StringLower		= (wchar_t*) malloc( len1 * sizeof( wchar_t ) );
	SubStringLower	= (wchar_t*) malloc( len2 * sizeof( wchar_t ) );

	if ( !StringLower || !SubStringLower ) {
		free( StringLower );
		free( SubStringLower );
		return NULL;
	}


	wcscpy_s( StringLower, len1, String );
	wcscpy_s( SubStringLower, len2, SubString );
	CharLowerW( StringLower );				//	_wcslwr_s( StringLower, len1 );
	CharLowerW( SubStringLower );			//	_wcslwr_s( SubStringLower, len2 );

	Found = wcsstr( StringLower, SubStringLower );

	if ( Found )
		Found = (wchar_t*) String + ( Found - StringLower );

	free( StringLower );
	free( SubStringLower );
	return Found;
}



DLLExports int f_expreplace( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t MainString[MAX_ARG_LEN]			= L"";
	wchar_t FromString[MAX_ARG_LEN]			= L"";
	wchar_t ToString[MAX_ARG_LEN]			= L"";
	wchar_t EscChars[SHORT_BUF_LEN]			= L"";
	wchar_t temp[MAX_ARG_LEN]				= L"";

	unsigned int Flags		= 0;
	unsigned int EscFlags	= 0;
	wchar_t *BeginAt		= NULL;
	wchar_t *Found			= NULL;
	wchar_t *Remainder		= NULL;
	size_t len				= 0;


	ParseArgs( String, PARSE_BREAK_COMMAS );


	if ( arg[0] ) {
		wcscpy_s( FromString, MAX_ARG_LEN, arg[0] );

		if ( ExpandEscapes( FromString, MAX_ARG_LEN ) == ESC_BUFFER_OVERFLOW ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Buffer overflow expanding \u2018from\u2019 string\r\n" );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing \u2018from\u2019 string\r\n" );
		return -1;
	}

	if ( arg[1] ) {
		wcscpy_s( ToString, MAX_ARG_LEN, arg[1] );

		if ( ExpandEscapes( ToString, MAX_ARG_LEN ) == ESC_BUFFER_OVERFLOW ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Buffer overflow expanding \u2018to\u2019 string\r\n" );
			return -1;
		}
	}

	if ( arg[2] ) {
		if ( ParseInt( arg[2], &Flags, NULL ) ) {
			_wcsupr_s( arg[2], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad flags: \"%s\"\r\n", arg[2] );
			return -1;
		}
	}

	if ( arg[3] ) {
		wcscpy_s( MainString, MAX_ARG_LEN, arg[3] );

		if ( ExpandEscapes( MainString, MAX_ARG_LEN ) == ESC_BUFFER_OVERFLOW ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Buffer overflow expanding main string\r\n" );
			return -1;
		}
	}

	if ( MainString[0] == '\0' ) {
		String[0] = '\0';
		return 0;
	}


	GetEscChars( EscChars, SHORT_BUF_LEN, &EscFlags );


	/*
	wprintf( L"   * MainString = '%s'\n", MainString );
	wprintf( L"   * FromString = '%s'\n", FromString );
	wprintf( L"   * ToString   = '%s'\n", ToString );
	wprintf( L"   * Flags      = %u\n", Flags );
	*/


	BeginAt = MainString;

	while ( *BeginAt ) {

		if ( Flags & 1 )
			Found = WcsIStr( BeginAt, FromString );
		else
			Found = wcsstr( BeginAt, FromString );

		if ( !Found )
			break;

		temp[0] = '\0';

		if ( Found > MainString ) 
			wcsncpy_s( temp, MainString, Found - MainString );

		wcscat_s( temp, MAX_ARG_LEN, ToString );
		len = wcslen( temp );

		Remainder = Found + wcslen( FromString );
		if ( *Remainder )
			wcscat_s( temp, MAX_ARG_LEN, Remainder );

		wcscpy_s( MainString, MAX_ARG_LEN, temp );
		BeginAt = MainString + len;

		if ( Flags & 2 )
			break;
	}


	if ( ( Flags & 4 ) == 0 ) {
		if ( Escapify( MainString, EscFlags, EscChars, MAX_ARG_LEN ) == ESC_BUFFER_OVERFLOW ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Buffer overflow escapifying resulting string\r\n" );
			return -1;
		}
	}

	wcscpy_s( String, MAX_ARG_LEN, MainString );
	return 0;
}



DLLExports int WINAPI _cescver( LPTSTR outString )
{
	if ( !outString )
		return 1;


	MultiByteToWideChar( CP_ACP, 0, MACRO_STR( EXPAND_ESCAPES_CPP ), -1, outString, MAX_ARG_LEN );
	return 0;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



DLLExports int expecho( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t *ch			= String;


	while ( iswspace( *ch ) )
		ch++;

	if ( !wcscmp( ch, L"/?" ) )
		return ShowCmdHelp( ExpEchoHelpText, FALSE );
	else if ( !wcscmp( ch, L"/??" ) )
		return ShowCmdHelp( ExpEchoHelpText, TRUE );


	ch = String;
	if ( iswspace( *ch ) )
		ch++;

	if ( ExpandEscapes( ch, MAX_ARG_LEN ) == ESC_BUFFER_OVERFLOW ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Buffer overflow\r\n" );
		return 2;
	}

	QPuts( ch );
	return 0;
}



#ifdef CESC_EXPTEST_CMD

DLLExports int exptest( LPTSTR String )
{
	if ( !String )
		return 1;


	DWORD CharStart[MAX_ARG_LEN];

	wchar_t *ch			= String;
	int rv				= 0;


	while ( iswspace( *ch ) )
		ch++;

	if ( !wcscmp( ch, L"/?" ) )
		return ShowCmdHelp( ExpTestHelpText, FALSE );
	else if ( !wcscmp( ch, L"/??" ) )
		return ShowCmdHelp( ExpTestHelpText, TRUE );


	ch = String;
	if ( iswspace( *ch ) )
		ch++;

	rv = ExpandEscapes( ch, MAX_ARG_LEN, CharStart );
	if ( rv < 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error %d\r\n", rv );
		return 2;
	}

	QPuts( ch );
	CrLf();

	for ( size_t j = 0; j <= wcslen( ch ); j++ )
		Printf( L"%3u:  %3u  %c\r\n", j, CharStart[j], ch[j] >= 0x20 ? ch[j] : 0x20 );

	return 0;
}

#endif	//	CESC_EXPTEST_CMD



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
			MyDLLHandle = (HINSTANCE) hModule;
			break;

		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/cesc.html";
	piInfo.pszDescription	= (LPTSTR) L"CEsc plugin";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( wcscmp( shortbuf, L"00" ) == 0 )
		Printf( L"%s v%u.%u.%u loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{

#ifdef NEWHELP_VERSION
	CloseHelpWindow();
#endif	//	NEWHELP_VERSION

	return 0;
}



const int NumHelpTexts							= 1;
const wchar_t HelpFeature[NumHelpTexts][15]		= { L"expecho" };
const wchar_t *HelpText[NumHelpTexts]			= { ExpEchoHelpText };




DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( ( inString[0] == '@' ) || ( inString[0] == '_' ) )
		return 0;

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			wcscpy_s( outString, SHORT_BUF_LEN, UsageText( HelpText[i] ) );
			return 1;
		}
	}

	return 0;
}

